from contextlib import AsyncExitStack, asynccontextmanager
from typing import AsyncGenerator, Dict, List, TypeVar

from ._mocked import Mocked

__all__ = ("stacked",)

MockedType = TypeVar("MockedType", List[Mocked], Dict[str, Mocked])


@asynccontextmanager
async def stacked(mocks: MockedType) -> AsyncGenerator[MockedType, None]:
    async with AsyncExitStack() as stack:
        if isinstance(mocks, list):
            yield [await stack.enter_async_context(mock) for mock in mocks]
        elif isinstance(mocks, dict):
            yield {name: await stack.enter_async_context(mock) for name, mock in mocks.items()}
        else:
            raise TypeError(f"Unsupported type: {type(mocks)}")



import mimetypes

from ._delayed_response import DelayedResponse
from ._relay_response import RelayResponse
from ._response import Response
from ._static_response import StaticResponse
from ._stream_response import StreamResponse

__all__ = ("Response", "StaticResponse", "StreamResponse", "RelayResponse", "DelayedResponse",)


mimetypes.add_type("application/json", ".json")



from typing import Any, Dict, Tuple
from urllib.parse import urljoin

from aiohttp import ClientSession
from aiohttp.typedefs import LooseHeaders
from aiohttp.web_request import BaseRequest
from multidict import CIMultiDict, CIMultiDictProxy
from packed import packable

from ._response import Response

__all__ = ("RelayResponse",)


# http://tools.ietf.org/html/rfc2616#section-13.5.1
_HOP_BY_HOP_HEADERS = (
    "connection",
    "keep-alive",
    "proxy-authenticate",
    "proxy-authorization",
    "te",
    "trailers",
    "transfer-encoding",
    "upgrade")

_FILTERED_HEADERS = _HOP_BY_HOP_HEADERS + ("host", "content-length")

_TargetResponseType = Tuple[int, Any, CIMultiDictProxy[str], bytes]


@packable("jj.responses.RelayResponse")
class RelayResponse(Response):
    def __init__(self, *, target: str) -> None:
        super().__init__()
        self._target = target
        self._prepare_hook_called = False

    @property
    def target(self) -> str:
        return self._target

    def _filter_headers(self, headers: LooseHeaders) -> CIMultiDict[str]:
        filtered: CIMultiDict[str] = CIMultiDict()
        for key, value in headers.items():
            if key.lower() not in _FILTERED_HEADERS:
                filtered[key] = value
        return filtered

    async def _do_target_request(self, request: BaseRequest) -> _TargetResponseType:
        url = urljoin(self._target, request.path)
        headers = self._filter_headers(request.headers)

        data = await request.read()

        async with ClientSession(auto_decompress=False) as session:
            async with session.request(request.method, url, params=request.query, headers=headers,
                                       data=data, allow_redirects=True) as response:
                body = await response.read()
        return response.status, response.reason, response.headers, body

    async def _prepare_hook(self, request: BaseRequest) -> "RelayResponse":
        if self._prepare_hook_called:
            return self
        status, reason, headers, body = await self._do_target_request(request)
        self.set_status(status, reason)
        self._headers = CIMultiDict(self._filter_headers(headers))
        self.body = body
        self._prepare_hook_called = True
        return self

    def copy(self) -> "RelayResponse":
        assert not self.prepared
        return self.__class__(target=self._target)

    def __packed__(self) -> Dict[str, Any]:
        return {"target": self._target}

    @classmethod
    def __unpacked__(cls, *, target: str, **kwargs: Any) -> "RelayResponse":  # type: ignore
        return cls(target=target)



import os
from typing import Optional, Union

from jj.expiration_policy import ExpirationPolicy
from jj.logs import SystemLogFilter  # backward compatibility
from jj.matchers import LogicalMatcher, RequestMatcher

from ._history import (
    HistoryAdapterType,
    HistoryFormatter,
    HistoryItem,
    HistoryRepository,
    HistoryRequest,
    HistoryResponse,
    PrettyHistoryFormatter,
    default_history_adapter,
)
from ._mock import Mock
from ._mocked import Mocked
from ._remote_handler import RemoteHandler
from ._remote_mock import RemoteMock
from ._remote_response import RemoteResponseType
from ._stacked import stacked

REMOTE_MOCK_URL = os.environ.get("JJ_REMOTE_MOCK_URL", "http://localhost:8080")
REMOTE_MOCK_DISPOSABLE = os.environ.get("JJ_REMOTE_MOCK_DISPOSABLE", "True")

REMOTE_MOCK_PPRINT = os.environ.get("JJ_REMOTE_MOCK_PPRINT", "False")
REMOTE_MOCK_PPRINT_LENGTH = os.environ.get("JJ_REMOTE_MOCK_PPRINT_LENGTH", "")
REMOTE_MOCK_PPRINT_WIDTH = os.environ.get("JJ_REMOTE_MOCK_PPRINT_WIDTH", "")

# backward compatibility
_REMOTE_MOCK_URL = REMOTE_MOCK_URL
_REMOTE_MOCK_DISPOSABLE = REMOTE_MOCK_DISPOSABLE


def mocked(matcher: Union[RequestMatcher, LogicalMatcher],
           response: RemoteResponseType,
           expiration_policy: Optional[ExpirationPolicy] = None,
           *,
           disposable: Optional[bool] = None,
           prefetch_history: bool = True,
           history_formatter: Optional[HistoryFormatter] = None,
           history_adapter: Optional[HistoryAdapterType] = default_history_adapter) -> "Mocked":
    if disposable is None:
        disposable = True if REMOTE_MOCK_DISPOSABLE.lower() == 'true' else False

    if (history_formatter is None) and (REMOTE_MOCK_PPRINT.lower() == 'true'):
        _history_width = int(REMOTE_MOCK_PPRINT_WIDTH) if REMOTE_MOCK_PPRINT_WIDTH else None
        _history_length = int(REMOTE_MOCK_PPRINT_LENGTH) if REMOTE_MOCK_PPRINT_LENGTH else None
        history_formatter = PrettyHistoryFormatter(_history_width, _history_length)

    handler = create_remote_handler(
        matcher, response, expiration_policy, history_adapter=history_adapter
    )

    return Mocked(handler, disposable=disposable, prefetch_history=prefetch_history,
                  history_formatter=history_formatter)


def create_remote_handler(matcher: Union[RequestMatcher, LogicalMatcher],
                          response: RemoteResponseType,
                          expiration_policy: Optional[ExpirationPolicy] = None,
                          *,
                          mock_url: str = REMOTE_MOCK_URL,
                          history_adapter: Optional[HistoryAdapterType] = default_history_adapter
                          ) -> RemoteHandler:
    return RemoteMock(mock_url).create_handler(matcher, response, expiration_policy,
                                               history_adapter=history_adapter)


__all__ = ("Mock", "mocked", "stacked", "create_remote_handler", "RemoteMock", "RemoteHandler",
           "Mocked", "HistoryRepository", "HistoryRequest", "HistoryResponse", "HistoryItem",
           "RemoteResponseType", "HistoryAdapterType", "SystemLogFilter",
           "default_history_adapter", "REMOTE_MOCK_URL", "REMOTE_MOCK_DISPOSABLE",)



import argparse

import jj
from jj.mock import Mock


def run() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("-H", "--host", type=str, default=None)
    parser.add_argument("-p", "--port", type=int, default=8080)
    args = parser.parse_args()

    jj.serve(Mock(), host=args.host, port=args.port)



import asyncio
import threading
from asyncio import Future
from typing import Any, Callable, Coroutine, TypeVar

__all__ = ("Thread",)

T = TypeVar("T")
TargetType = Callable[..., Coroutine[None, None, T]]


class Thread(threading.Thread):
    def __init__(self, future: "Future[T]", target: TargetType[T],
                 *args: Any, **kwargs: Any) -> None:
        super().__init__()
        self._future = future
        self._target = target
        self._args = args
        self._kwargs = kwargs

    def run(self) -> None:
        try:
            result = asyncio.run(self._target(*self._args, **self._kwargs))
        except BaseException as e:
            self._future.set_exception(e)
        else:
            self._future.set_result(result)


def run_async(target: TargetType[T], *args: Any, **kwargs: Any) -> T:
    try:
        asyncio.get_running_loop()
    except RuntimeError:
        return asyncio.run(target(*args, **kwargs))
    else:
        future: "Future[T]" = Future()
        thread = Thread(future, target, *args, **kwargs)
        thread.start()
        thread.join()
        return future.result()



from http.cookies import Morsel
from typing import Any, Dict, Union

from aiohttp.payload import BytesPayload, IOBasePayload, TextIOPayload

__all__ = ("cookie_to_dict", "get_response_body",)


def cookie_to_dict(cookie: "Morsel[str]") -> Dict[str, Union[str, None]]:
    dictionary: Dict[str, Union[str, None]] = {
        "name": cookie.key,
        "value": cookie.value,
    }
    for attr in ("expires", "domain", "max-age", "path", "secure", "httponly", "version"):
        key = attr.replace("-", "_")
        val = cookie.get(attr)
        dictionary[key] = None if val == "" else val
    return dictionary


def get_response_body(body: Any) -> bytes:
    if isinstance(body, (bytes, bytearray, memoryview)):
        return bytes(body)
    elif isinstance(body, BytesPayload):
        return bytes(body._value)
    elif isinstance(body, TextIOPayload):
        return bytes(body._value.read(), body.encoding)  # type: ignore
    elif isinstance(body, IOBasePayload):
        return bytes(body._value.read())
    else:
        raise ValueError("Unsupported body type {}".format(type(body)))



from setuptools import find_packages, setup


def find_required():
    with open("requirements.txt") as f:
        return f.read().splitlines()


def find_dev_required():
    with open("requirements-dev.txt") as f:
        return f.read().splitlines()


setup(
    name="jj",
    version="2.10.3",
    description="Remote HTTP Mock",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Nikita Tsvetkov",
    author_email="tsv1@fastmail.com",
    python_requires=">=3.8",
    url="https://github.com/jj-mock/jj",
    project_urls={
        "Docs": "https://jj-mock.io",
        "GitHub": "https://github.com/jj-mock/jj",
    },
    license="Apache-2.0",
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_data={"jj": ["py.typed"]},
    entry_points={
        "console_scripts": ["jj = jj._entry_point:run"]
    },
    install_requires=find_required(),
    tests_require=find_dev_required(),
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Typing :: Typed",
    ],
)
